#############################################################
#
# dbccp4i_test_client.tcl
#
# Client application that tests the various database functions
# via the Tcl client API.
#
# CVS_Id $Id: dbccp4i_test_client.tcl,v 1.4 2008/01/08 15:35:45 pjx Exp $
#
#############################################################

# Announce yourself
puts "=========================================================="
puts " dbccp4i_test_client"
puts " Test database functions via the Tcl client API"
puts "==========================================================\n"

# Acquire the handler client API functions
source [file join $env(DBCCP4I_TOP) ClientAPI dbClientAPI.tcl]
namespace import dbClientAPI::*

# Procedures
proc exit_test { status } {
    # Do cleanup and exit
    global test_project_dir
    # Shut down the handler
    if { [DbVerifyConnection] } {
	puts -nonewline "Issuing the shutdown command to the handler..."
	DbShutdown
	puts "done"
    }
    # Pause for a second
    after 1000
    # Remove the test directory from this run
    if { [file exists $test_project_dir] &&  [file isdirectory $test_project_dir] } {
	puts -nonewline "Removing test directory $test_project_dir..."
	file delete -force $test_project_dir
	puts "done"
    }
    puts ""
    puts "=========================================================="
    puts -nonewline " dbccp4i_test_client: finished"
    if { ! $status } {
	puts " (OK)"
    } else {
	puts " (ERROR)"
    }
    puts "=========================================================="
    exit $status
}

# Set the debugging output level
DbXMLSetDebug 0

# Connect to the handler
puts -nonewline "Connecting to handler..."
if { ![DbHandlerConnect dbccp4i_test_client] } {
    puts "failed: couldn't connect to the handler"
    exit 1
}
puts "done"

# Report settings
puts "Handler information:"
puts "\t[DbInfo]"
puts "Supported database backends:"
puts "\tDef file: [DbSupported def]"
puts "\tSqlite  : [DbSupported sqlite]"

# List projects
puts -nonewline "Acquiring listing of projects..."
if { [catch { set projects [ListProjects] } err] } {
    puts "failed with error \"$err\""
    exit 1
}
puts "done"

# Check that data is consistent
puts -nonewline "Acquiring NProjects..."
if { [catch { set nprojects [GetNProjects] } err] } {
    puts "failed with error \"$err\""
    exit 1
}
puts "done: $nprojects"
if { [llength $projects] != $nprojects } {
    puts "*** NProjects inconsistent with number of projects listed ***"
    exit 1
}

# Generate a unique project name for the tests
set test_project "TEST_[clock format [clock seconds] -format %G%j%H%M%S]"
set test_project_dir [file join $env(CCP4_SCR) $test_project]
puts "Test project name: $test_project"
puts "Test project dir : $test_project_dir"

# Check whether test project name already exists
if { [lsearch $projects $test_project] > -1 } {
    puts "Test project name already in use"
    exit 1
}

# Check whether test project directory already exists
if { [file exists $test_project_dir] } {
    puts "*** Directory $test_project_dir already exists ***"
    exit_test 1
}

# Create the test project
puts -nonewline "Creating test project $test_project..."
if { [catch { set status [CreateProject $test_project $test_project_dir msg] } err] } {
    puts "failed: error message \"$err\""
    exit 1
} else {
    if { ! $status } {
	puts "failed: API returned \"$msg\""
	exit_test 1
    } else {
	puts "done"
    }
}
# Verify that the project exists
if { [lsearch [ListProjects] $test_project] < 0 } {
    puts "*** Test project not in the list of projects ***"
    exit_test 1
}

# Add a new job
puts -nonewline "Creating new job in project $test_project..."
if { [catch { set jobid [NewJob $test_project] } err] } {
    puts "failed: error message \"$err\""
    exit 1
} else {
    if { $jobid < 1 } {
	puts "failed: API returned job id as $jobid"
	exit_test 1
    } else {
	puts "done"
    }
}

# Remove the test project
puts -nonewline "Deleting test project $test_project..."
if { [catch { set status [DeleteProject $test_project] } err] } {
    puts "failed: error message \"$err\""
    exit_test 1
} else {
    if { ! $status } {
	puts "failed: API returned \"$msg\""
	exit_test 1
    } else {
	puts "done"
    }
}

# Verify that the project has been removed
if { [lsearch [ListProjects] $test_project] > -1 } {
    puts "*** Test project still in the list of projects ***"
    exit_test 1
}

# Finished
puts -nonewline "Disconnecting from handler..."
if { [catch { DbHandlerDisconnect } err] } {
    puts "failed: disconnection error \"$err\""
    exit_test 1
}
puts "done"

# Exit everything ok
exit_test 0